#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.icons.font_awesome
      @file: form_icons.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.modules.cli.icons.font_awesome.awesome import Awesome


class FormIcons(Awesome):
    """
        Form UI icons.
        Codes can be found here:
        - https://fontawesome.com/cheatsheet?from=io
    """

    # @formatter:off
    ARROW_LEFT = '\uF060'       # 
    ARROW_RIGHT = '\uF061'      # 
    ARROW_UP = '\uF062'         # 
    ARROW_DOWN = '\uF063'       # 
    MOVE_ARROWS = '\uF047'      # 
    SELECTOR = '\uFC32'         # ﰲ

    CHECK = '\uF00C'            # 
    UNCHECK = '\uF00D'          # 
    PLUS_CIRCLE = '\uF055'      # 
    MINUS_CIRCLE = '\uF056'     # 
    ERROR_CIRCLE = '\uf057'     # 
    CHECK_CIRCLE = '\uF058'     # 
    QUESTION_CIRCLE = '\uF059'  # 
    UNCHECK_CIRCLE = '\uF111'   # 
    CHECK_SQUARE = '\uF14A'     # 
    UNCHECK_SQUARE = '\uF0C8'   # 

    ON = '\uF205'               # 
    OFF = '\uF204'              # 
    MARKED = '\uF634'           # 
    UNMARKED = '\uF630'         # 
    CLEAR = '\uF5E1'            # 
    REFRESH = '\uF01E'          # 
    FILTER  = '\uF0B0'          # 
    ATTACH = '\uF0C6'           # 

    HIDDEN = '\uF070'           # 
    VISIBLE = '\uF06E'          # 
    LOCKED = '\uF023'           # 
    UNLOCKED = '\uF09C'         # 
    EDITABLE = '\uF044'         # 
    MASKED = '\uF0CE'           # 
    SELECTABLE = '\uF150'       # 
    DESELECT = '\uF657'         # 

    DELETE = '\uF014'           # 
    EDIT = '\uF040'             # 
    PLUS = '\uF067'             # 
    MINUS = '\uF068'            # 
    # @formatter:on


if __name__ == '__main__':
    FormIcons.demo_icons()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @Package: main.enums
      @file: exit_status.py
   @created: Thu, 20 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.core.enums.enumeration import Enumeration


class ExitStatus(Enumeration):
    """The exit status of an executed command is the value returned by the waitpid system call or equivalent function.
    Exit statuses fall between 0 and 255, though, as explained below, the shell may use values above 125 specially.
    Exit statuses from shell builtins and compound commands are also limited to this range. Under certain circumstances,
    the shell will use special values to indicate specific failure modes."""

    # @formatter:off

    # Returned when something went wrong due to any Human interaction
    ERROR   = 129

    # Returned when something ran successfully without errors
    SUCCESS =  0

    # Returned when something that was supposed to work and failed due to unexpected software behaviour
    FAILED  =  1

    # Returned when something ran should be aborted. Generally due to usr cancellation
    ABORTED =  2

    # @formatter:on

    def __str__(self):
        return self.name

    def __repr__(self):
        return "{}{}({})".format(
            '\033[0;32m' if self.name == 'SUCCESS' else '\033[0;31m',
            self.name, self.value
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.hspylib.core.datasource.cassandra
      @file: cassandra_configuration.py
   @created: Sat, 12 Nov 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from typing import Optional

from hspylib.core.datasource.db_configuration import DBConfiguration


class CassandraConfiguration(DBConfiguration):
    """Represents a Cassandra datasource configuration"""

    def __init__(
        self,
        resource_dir: str,
        filename: Optional[str] = None,
        profile: Optional[str] = None):
        super().__init__(resource_dir, filename, profile)
        self._secure_bundle_path = self['datasource.secure_bundle_path']
        self._protocol_version = self['datasource.protocol_version'] or 4
        self._connect_timeout = self['datasource.connect_timeout'] or 30
        self._default_timeout = self['datasource.default_timeout'] or 60

    @property
    def secure_bundle_path(self) -> str:
        return self._secure_bundle_path

    @property
    def protocol_version(self) -> int:
        return self._protocol_version

    @property
    def connect_timeout(self) -> int:
        return self._connect_timeout

    @property
    def default_timeout(self) -> int:
        return self._default_timeout



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.application
      @file: options_builder.py
   @created: hu, 14 Feb 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from argparse import ArgumentParser
from typing import Any, Union

from hspylib.modules.cli.application.parser_action import ParserAction


class OptionsBuilder:
    """TODO"""

    def __init__(self, arg_parser: ArgumentParser):
        self._arg_parser = arg_parser

    def option(
        self,
        name: str,
        shortopt: str,
        longopt: str,
        help_string: str = None,
        choices: list = None,
        required: bool = False,
        nargs: Union[str, int] = None,
        default: Any = None) -> 'OptionsBuilder':
        """TODO"""

        self._arg_parser.add_argument(
            f"-{shortopt.replace('^-', '')[0]}",
            f"--{longopt.replace('^-*', '')}",
            dest=name,
            help=help_string or f'the {longopt}',
            required=required,
            action=ParserAction.STORE.value,
            choices=choices,
            nargs=nargs,
            default=default)

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.hspylib.core.datasource.postgres
      @file: postgres_repository.py
   @created: Sat, 12 Nov 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import contextlib
import logging as log
from abc import abstractmethod
from typing import Generic, List, Optional, Tuple, TypeVar

import redis
from redis.client import Pipeline
from retry import retry

from hspylib.core.datasource.crud_entity import CrudEntity
from hspylib.core.datasource.db_repository import Connection, Cursor
from hspylib.core.datasource.redis.redis_configuration import RedisConfiguration
from hspylib.core.enums.charset import Charset
from hspylib.core.exception.exceptions import DatabaseConnectionError, DatabaseError
from hspylib.core.preconditions import check_not_none

T = TypeVar('T', bound=CrudEntity)


class RedisRepository(Generic[T]):
    """Implementation of a data access layer for a postgres persistence store.
    Ref.: https://github.com/redis/redis-py
    Ref.: https://docs.redis.com/latest/rs/references/client_references/client_python/
    """

    def __init__(self, config: RedisConfiguration):
        self._config = config

    def __str__(self):
        return f"{self.hostname}:{self.port}/{self.database}"

    def __repr__(self):
        return str(self)

    @property
    def logname(self) -> str:
        """TODO"""
        return self.__class__.__name__.split('_')[0]

    @property
    def hostname(self) -> str:
        return self._config.hostname

    @property
    def port(self) -> int:
        return self._config.port

    @property
    def database(self) -> str:
        return self._config.database

    @property
    def password(self) -> str:
        return self._config.password

    @property
    def ssl(self) -> bool:
        return self._config.ssl

    @retry(tries=3, delay=2, backoff=3, max_delay=30)
    def _create_session(self) -> Tuple[Connection, Cursor]:
        """TODO"""
        log.debug(f"{self.logname} Attempt to connect to database: {str(self)}")
        conn = redis.Redis(
            ssl=self.ssl,
            host=self.hostname,
            port=self.port,
            password=self.password)
        log.debug(f"{self.logname} Connection info: {conn.config_get('databases')}")
        return conn, conn.pipeline()

    @contextlib.contextmanager
    def pipeline(self) -> Pipeline:
        """TODO"""
        conn, pipe = None, None
        try:
            conn, pipe = self._create_session()
            log.debug(f"{self.logname} Successfully connected to database: {str(self)} [ssid={hash(pipe)}]")
            yield pipe
        except redis.exceptions.ConnectionError as err:
            raise DatabaseConnectionError(f"Unable to open/execute-on database session => {err}") from err
        except redis.exceptions.ResponseError as err:
            log.error(f"{self.logname} Pipeline failed with: {err}")
        except Exception as err:
            raise DatabaseError(f"{self.logname} Unable to execute pipeline -> {err}") from err
        finally:
            if pipe:
                pipe.close()

    def delete(self, *keys: str) -> int:
        """TODO"""
        check_not_none(keys)
        with self.pipeline() as pipe:
            pipe.delete(*keys)
            ret_val = pipe.execute() or []
            log.debug(f"{self.logname} "
                      f"Executed a pipelined 'DEL' command and returned: {ret_val}")
            return ret_val[0] or 0

    def get(self, *keys: str) -> List[T]:
        """TODO"""
        check_not_none(keys)
        with self.pipeline() as pipe:
            result = []
            pipe.mget(keys)
            count = len(pipe)
            ret_val = list(filter(None, pipe.execute()))
            if ret_val:
                list(map(lambda e: result.append(self.to_entity_type(e)), ret_val[0]))
            log.debug(f"{self.logname} "
                      f"Executed '{count}' pipelined 'GET' command(s) and returned {len(result)} entries")
            return result

    def get_one(self, key: str) -> Optional[T]:
        """TODO"""
        check_not_none(key)
        with self.pipeline() as pipe:
            pipe.get(key)
            ret_val = list(filter(None, pipe.execute()))
            log.debug(f"{self.logname} "
                      f"Executed a pipelined 'GET' command and returned: {ret_val}")
            return self.to_entity_type(ret_val[0]) if ret_val else None

    def set(self, *entities: T) -> None:
        """TODO"""
        check_not_none(entities)
        with self.pipeline() as pipe:
            list(map(lambda e: pipe.set(self.build_key(e), str(e.as_dict())), entities))
            count = len(pipe)
            pipe.execute()
            log.debug(f"{self.logname} "
                      f"Executed '{count}' pipelined 'SET' command(s)")

    def flush_all(self) -> None:
        """TODO"""
        with self.pipeline() as pipe:
            pipe.flushall()
            pipe.execute()
            log.debug(f"{self.logname} "
                      f"Executed a FLUSHALL command")

    def keys(self, pattern: str) -> List[str]:
        """TODO"""
        with self.pipeline() as pipe:
            pipe.keys(pattern)
            ret_val = list(filter(None, pipe.execute()))
            result = list(map(lambda k: str(k, Charset.UTF_8.value), ret_val[0])) if ret_val else []
            log.debug(f"{self.logname} "
                      f"Executed a KEYS command and returned: {len(ret_val)} entries")
            return result

    @abstractmethod
    def build_key(self, entity: T) -> str:
        """TODO"""

    @abstractmethod
    def to_entity_type(self, entity_string: bytes) -> T:
        """TODO"""



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.123
#
# Package: main.hspylib.core.datasource
"""Package initialization."""

__all__ = [
  'cassandra', 
  'crud_entity', 
  'crud_repository', 
  'crud_service', 
  'db_configuration', 
  'db_repository', 
  'firebase', 
  'identity', 
  'mysql', 
  'postgres', 
  'redis', 
  'sqlite'
]
__version__ = '0.11.123'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.hspylib.core.datasource
      @file: identity.py
   @created: Thu, 03 Nov 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import random
import string
import sys
from collections import namedtuple
from typing import Tuple, Type, Union
from uuid import UUID, uuid4

from hspylib.core.preconditions import check_argument
from hspylib.core.tools.namespace import Namespace

IDENTITY = Union[Tuple, int, str, UUID]


class Identity(Namespace):

    @classmethod
    def auto(cls, field_name: str = 'id', id_type: Type = UUID) -> 'Identity':
        _id_ = namedtuple('Identity', [f"{field_name}"])
        match id_type.__name__:
            case 'int':
                number = random.randint(0, sys.maxsize)
                return Identity(_id_(number))
            case 'str':
                letters = string.ascii_lowercase + '-_'
                st = ''.join(random.choice(letters) for _ in range(32))
                return Identity(_id_(st))
            case 'UUID':
                return Identity(_id_(uuid4().hex))
            case other:
                raise NotImplementedError(f"auto-identity generator for type '{other}' is not implemented")

    def __init__(self, identity: Tuple | 'Identity'):
        check_argument(isinstance(identity, Tuple | Identity), "Must be a named tuple or Identity")
        identity = identity.identity if isinstance(identity, Identity) else identity
        self._identity = identity
        super().__init__('Identity', **self.as_dict())

    def __str__(self) -> str:
        return str(self._identity)

    @property
    def identity(self) -> Tuple:
        return self._identity

    def as_dict(self) -> dict:
        return self._identity._asdict()

    def as_column_set(self, separator: str = ',') -> str:
        """TODO"""
        column_set = []
        list(map(
            lambda key, value: column_set.append(f"{key} = {value}"), self.attributes, self.values
        ))
        return separator.join(column_set)


if __name__ == '__main__':
    PersonId = namedtuple('PersonId', ['uuid'])
    UserId = namedtuple('UserId', ['uid', 'email'])
    i1 = Identity.auto('uid')
    i2 = Identity(i1)
    i3 = Identity(UserId('12345', 'user@example.com'))
    i4 = Identity(PersonId(uuid4().hex))
    print('Identities: ', i1, i2, i3, i4)
    print('Values: ', i1.values, i2.values, i3.values, i4.values)
    print('Items: ', i1.items(), i2.items(), i3.items(), i4.items())
    print('Attrs/Values: ', i1.attributes, i1.values)
    print('Attrs/Values: ', i2.attributes, i2.values)
    print('Attrs/Values: ', i3.attributes, i3.values)
    print('Attrs/Values: ', i4.attributes, i4.values)
    print('As Dict: ', i1.as_dict(), i2.as_dict(), i3.as_dict(), i4.as_dict())
    print('As ColSet: ', i1.as_column_set(), i2.as_column_set(), i3.as_column_set(), i4.as_column_set())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.hspylib.core.datasource.firebase
      @file: firebase_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import json
import logging as log
from abc import abstractmethod
from typing import Any, Generic, List, Optional, TypeVar

from requests.exceptions import HTTPError

from hspylib.core.datasource.crud_entity import CrudEntity
from hspylib.core.datasource.firebase.firebase_configuration import FirebaseConfiguration
from hspylib.core.datasource.identity import Identity
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.preconditions import check_not_none
from hspylib.modules.fetch.fetch import delete, get, put
from hspylib.modules.fetch.http_response import HttpResponse

T = TypeVar('T', bound=CrudEntity)


class FirebaseRepository(Generic[T]):
    """Implementation of a data access layer for a Firebase persistence store.
    API   Ref.: https://firebase.google.com/docs/reference/rest/database
    Auth  Ref.: https://firebase.google.com/docs/database/rest/auth#python
    Order Ref:. https://firebase.google.com/docs/database/rest/retrieve-data#section-rest-ordered-data
    """

    @staticmethod
    def _assert_response(response: HttpResponse, message: str) -> HttpResponse:
        check_not_none(response, "Response is none")
        if response.status_code not in [HttpCode.OK, HttpCode.ACCEPTED, HttpCode.CREATED]:
            raise HTTPError(f"{response.status_code} => {message}")
        return response

    @staticmethod
    def quote(value: Any) -> str:
        """Quote or double quote the value according to the value type. """
        if isinstance(value, bool):
            return f'{str(value).lower()}'
        elif isinstance(value, int | float):
            return str(value)
        else:
            return f'"{value}"' if value.startswith('\'') and value.endswith('\'') else f"'{value}'"

    def __init__(self, config: FirebaseConfiguration):
        self._payload = None
        self._config = config

    def __str__(self) -> str:
        return str(self._config) + f"TABLE_NAME={self.table_name() or ''}"

    def __repr__(self) -> str:
        return str(self)

    @property
    def logname(self) -> str:
        """TODO"""
        return self.__class__.__name__.split('_')[0]

    @property
    def config(self) -> FirebaseConfiguration:
        return self._config

    @property
    def info(self) -> str:
        return f"{self.config}"

    @property
    def hostname(self) -> str:
        return self._config.hostname

    @property
    def port(self) -> int:
        return self._config.port

    @property
    def base_url(self) -> str:
        return self._config.base_url

    @property
    def database(self) -> str:
        """Return the database name."""
        return self._config.database

    def delete(self, entity: T) -> None:
        """Deletes the given entity from the Firebase store"""
        self.delete_by_id(entity.identity)

    def delete_by_id(self, entity_id: Identity) -> None:
        """Deletes the given entity by it's Id from the Firebase store"""
        ids = '.'.join(entity_id.values)
        url = f'{self._config.url(self.table_name())}/{ids}.json'
        log.debug('Deleting firebase entry: \n\t|-Id=%s\n\t|-From %s', entity_id, url)
        self._assert_response(delete(url), f'Unable to delete from={url}')

    def delete_all(self, entities: List[T]) -> None:
        """Delete from Firebase all entries provided.
        TODO Check if there is a better ways of doing it.
        """
        list(map(self.delete, entities))

    def save(self, entity: T) -> None:
        """Saves the given entity at the Firebase store"""
        ids = '.'.join(entity.identity.values)
        url = f'{self._config.url(self.table_name())}/{ids}.json'
        payload = entity.as_json()
        log.debug("Saving firebase entry: \n\t|-%s \n\t|-Into %s", entity, url)
        self._assert_response(put(url, payload), f"Unable to put into={url} with json_string={payload}")

    def save_all(self, entities: List[T]) -> None:
        """Save from Firebase all entries provided.
        TODO Check if there is a better ways of doing it.
        """
        list(map(self.save, entities))

    def find_all(
        self,
        order_by: Optional[List[str]] = None,
        limit_to_first: Optional[int] = None,
        limit_to_last: Optional[int] = None,
        start_at: Optional[int | str] = None,
        end_at: Optional[int | str] = None,
        equal_to: Optional[int | str] = None) -> List[T]:
        """Return filtered entries from the Firebase store"""

        f_order_by = "orderBy=" + (','.join([f'"{o}"' for o in order_by]) if order_by else '"$key"')
        f_start_at = f"&startAt={self.quote(start_at)}" if start_at else ''
        f_end_at = f"&endAt={self.quote(end_at)}" if end_at else ''
        f_equal_to = f"&equalTo={self.quote(equal_to)}" if equal_to else ''
        f_limit_first = f"&limitToFirst={limit_to_first}" if limit_to_first else ''
        f_limit_last = f"&limitToLast={limit_to_last}" if limit_to_last else ''
        url = f'{self._config.url(self.table_name())}.json?' \
              f'{f_order_by}{f_start_at}{f_end_at}{f_equal_to}{f_limit_first}{f_limit_last}'
        log.debug('Fetching firebase entries: \n\t|-From %s', url)
        response = self._assert_response(get(url), f"Unable to get from={url}")
        if response.body and response.body != 'null':
            return self.to_entity_list(response.body)

        return []

    def find_by_id(self, entity_id: Identity) -> Optional[T]:
        """Return the entry specified by ID from the Firebase store, None if no such entry is found."""
        ids = '.'.join(entity_id.values)
        url = f'{self._config.url(self.table_name())}/{ids}.json'
        log.debug('Fetching firebase entry: \n\t|-Id=%s\n\t|-From %s', entity_id, url)
        response = self._assert_response(get(url), f"Unable to get from={url}")
        if response.body and response.body != 'null':
            return self.to_entity_type(json.loads(response.body))

        return None

    def exists_by_id(self, entity_id: Identity) -> bool:
        """Check if provided entity exists on the database.
        TODO Check if there is a better ways of doing it.
        """
        return self.find_by_id(entity_id) is not None

    @abstractmethod
    def to_entity_type(self, entity_dict: dict | tuple) -> T:
        """TODO"""

    @abstractmethod
    def table_name(self) -> str:
        """TODO"""

    def to_entity_list(self, json_string: str) -> List[T]:
        """Return filtered entries from the json_string as a list"""
        if json_string and (entities := json.loads(json_string)):
            ret_list = []
            list(map(lambda v: ret_list.append(self.to_entity_type(v)), entities.values()))
            return ret_list
        return []



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   main.modules.cli.tui.extra.minput
      @file: input_validator.py
   @created: Thu, 20 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import re
import typing


class InputValidator:
    """TODO"""

    RE_FMT_LETTER = r'^[a-zA-Z]{%min%,%max%}$'
    RE_FMT_WORD = r'^[a-zA-Z0-9 _]{%min%,%max%}$'
    RE_FMT_NUMBER = r'^[0-9\.]{%min%,%max%}$'
    RE_FMT_TOKEN = r'^\<?[a-zA-Z0-9_\- ]+\>?(\|\<?[a-zA-Z0-9_\- ]+\>?)*$'
    RE_FMT_MASKED = r'.*\|.+'
    RE_FMT_ANYTHING = r'^.{%min%,%max%}$'

    def __init__(self, min_length: int = 1, max_length: int = 30, pattern: str = None):
        self._min_length = min_length
        self._max_length = max_length
        self._pattern = pattern or self.RE_FMT_ANYTHING

    def __str__(self) -> str:
        return f"r\"{self._pattern}\""

    def __repr__(self):
        return str(self)

    @staticmethod
    def letters(min_length: int = 1, max_length: int = 30) -> typing.Any:
        return InputValidator(min_length, max_length, InputValidator.RE_FMT_LETTER)

    @staticmethod
    def words(min_length: int = 1, max_length: int = 30) -> typing.Any:
        return InputValidator(min_length, max_length, InputValidator.RE_FMT_WORD)

    @staticmethod
    def numbers(min_length: int = 1, max_length: int = 30) -> typing.Any:
        return InputValidator(min_length, max_length, InputValidator.RE_FMT_NUMBER)

    @staticmethod
    def anything(min_length: int = 1, max_length: int = 30) -> typing.Any:
        return InputValidator(min_length, max_length, InputValidator.RE_FMT_ANYTHING)

    @staticmethod
    def custom(pattern: str) -> typing.Any:
        return InputValidator(pattern=pattern)

    def validate(self, value: str) -> bool:
        regex = self._get_pattern()
        return re.search(regex, value) is not None

    def _get_pattern(self) -> str:
        return self._pattern \
            .replace('%min%', str(self._min_length or 1)) \
            .replace('%max%', str(self._max_length or 30))



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.123
#
# Package: main.hspylib.core.datasource.redis
"""Package initialization."""

__all__ = [
  'redis_configuration', 
  'redis_repository'
]
__version__ = '0.11.123'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.application
      @file: version.py
   @created: Thu, 14 Feb 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import logging as log
import os
import re
from pathlib import Path
from typing import Tuple, Union

from hspylib.core.enums.charset import Charset


class Version:

    @staticmethod
    def load(filename: str = '.version', load_dir: Union[str, Path] = os.getcwd()) -> 'Version':
        """Load a version from file in the form: [major.minor.build] ."""
        filepath = f'{str(load_dir)}/{filename}'
        if not os.path.exists(filepath):
            log.warning('File "%s" does not exist. Could not fetch application version', filepath)
            return Version((-1, -1, -1))

        return Version(Version.read(filepath))

    @staticmethod
    def read(filepath: str = ".version") -> Tuple[int, int, int]:
        """Retrieve the version from the version file in the form: [major.minor.build] ."""
        try:
            log.debug("Reading version from %s", filepath)
            with open(filepath, encoding=Charset.UTF_8.value) as fh_version:
                ver = tuple(map(int, map(str.strip, fh_version.read().split('.'))))
                return ver if ver and re.search(r'(\d+, \d+, \d+)', str(ver)) else (-1, -1, -1)
        except FileNotFoundError:
            return -1, -1, -1

    def __init__(self, version: Tuple[int, int, int]):
        self._version = version
        self._version_string = '.'.join(map(str, version))

    def __str__(self):
        return self._version_string

    def __repr__(self):
        return str(self)

    @property
    def version(self) -> str:
        """TODO"""
        return f'v{str(self)}'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.tui
      @file: menu_ui.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import signal
from typing import Optional

from hspylib.modules.cli.tui.menu.menu import Menu
from hspylib.modules.cli.vt100.vt_utils import exit_app


class MenuUi:
    """TODO"""

    def __init__(self, root: Optional[Menu]):
        self.done = False
        self.previous = None
        self.current = root
        self.next = None
        signal.signal(signal.SIGINT, exit_app)

    def show(self) -> None:
        """TODO"""

        while not self.done:
            if self.current:
                self.next = self.current.execute()
                if self.next:
                    self.change_menu(self.next)
                else:
                    self.done = True
            else:
                self.done = True

    def change_menu(self, menu: Menu) -> None:
        """TODO"""

        self.previous = self.current
        self.current = menu



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: hspylib
   @package: main.hspylib
      @file: __main__.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import sys

from hspylib.__classpath__ import _Classpath
from hspylib.addons.appman.appman import AppManager
from hspylib.addons.appman.appman_enums import AppType, Extension
from hspylib.addons.widman.widman import WidgetManager
from hspylib.core.enums.charset import Charset
from hspylib.core.enums.enumeration import Enumeration
from hspylib.core.tools.commons import run_dir, syserr
from hspylib.core.tools.text_tools import strip_linebreaks
from hspylib.modules.cli.application.application import Application
from hspylib.modules.cli.application.version import Version


class Main(Application):
    """HSPyLib Manager - Manage HSPyLib applications."""

    # The welcome message
    DESCRIPTION = _Classpath.get_source_path("welcome.txt").read_text(encoding=str(Charset.UTF_8))

    # Location of the .version file
    VERSION_DIR = _Classpath.source_root()

    class Addon(Enumeration):
        """HsPyLib addon types."""
        # @formatter:off
        APPMAN  = 'appman'
        WIDGETS = 'widgets'
        # @formatter:on

    def __init__(self, app_name: str):
        version = Version.load(load_dir=self.VERSION_DIR)
        super().__init__(app_name, version, self.DESCRIPTION.format(version))

    def _setup_arguments(self) -> None:
        """Initialize application parameters and options"""

        # @formatter:off
        self._with_chained_args('application', 'the HSPyLib application to run') \
            .argument(self.Addon.APPMAN.value, 'app Application Manager: Create HSPyLib based python applications') \
                .add_option(
                    'dest-dir', 'd', 'dest-dir',
                    'the destination directory. If omitted, the current directory will be used.',
                    nargs='?', default=self._run_dir) \
                .add_argument(
                    'app-name',
                    'the application name', nargs='?') \
                .add_argument(
                    'app-type',
                    'the application type. Appman is going to scaffold a basic application based on the app type',
                    choices=[
                        AppType.APP.value, AppType.QT_APP.value, AppType.WIDGET.value
                    ], nargs='?') \
                .add_argument(
                    'app-ext',
                    '"gradle" is going to initialize you project with gradle (requires gradle). '
                    '"git" is going to initialize a git repository (requires git)',
                    nargs='*') \
            .argument(self.Addon.WIDGETS.value, 'app Widgets Manager: Execute an HSPyLib widget') \
                .add_argument(
                    'widget-name',
                    'the name of the widget to be executed. If omitted, all available widgets will be '
                    'presented in a dashboard',
                    nargs='?') \
                .add_argument(
                    'widget-args', "the widget's arguments (if applicable)",
                    nargs='*') \
            # @formatter:on

    def _main(self, *params, **kwargs) -> int:
        """Main entry point handler"""
        self._run()
        return 0

    def _run(self) -> None:
        """Execute the application"""

        app = self.get_arg('application')
        if app == self.Addon.APPMAN.value:
            addon = AppManager(self)
            app_type = self.get_arg('app-type')
            if app_type:
                app_ext = self.get_arg('app-ext')
                addon.create(
                    self.get_arg('app-name'), AppType.of_value(app_type),
                    list(map(Extension.value_of, app_ext)) if app_ext else [],
                    self.get_arg('dest-dir') or run_dir())
            else:
                args = addon.prompt()
                if args:
                    app_ext = []
                    if args.initialize_gradle:
                        app_ext.append(Extension.GRADLE)
                    if args.initialize_git:
                        app_ext.append(Extension.GIT)
                    addon.create(
                        args.app_name, AppType.of_value(args.app_type),
                        list(app_ext),
                        args.dest_dir or run_dir())
        elif app == self.Addon.WIDGETS.value:
            addon = WidgetManager(self)
            widget_name = self.get_arg('widget-name')
            if widget_name:
                widget_args = list(map(strip_linebreaks, self.get_arg('widget-args')))
                addon.execute(widget_name, widget_args)
            else:
                addon.dashboard()
        else:
            syserr(f'### Invalid application: {app}')
            self.usage(1)


# Application entry point
if __name__ == "__main__":
    Main('hspylib').INSTANCE.run(sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: main.hspylib.core.decorator
      @file: decorators.py
   @created: Tue, 11 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import logging as log
import os
import unittest
from datetime import datetime, timedelta
from math import ceil
from typing import Any, Callable

from hspylib.core.tools.commons import str_to_bool


def integration_test(cls: type):
    """Enable/Disable HsPyLib integration tests"""

    it_disabled = str_to_bool(os.environ.get('HSPYLIB_IT_DISABLED', 'True'))
    if it_disabled:
        log.debug('Skipping test: ', cls.__name__)
        return unittest.skipIf(
            it_disabled,
            f'Disabled = {it_disabled} :integration tests because it needs real servers running'
        )

    return cls


def temporized(func: Callable):
    """Log the time spent on the Callable invocation."""

    def timedelta_to_str(delta: timedelta) -> str:
        """Convert timedelta object into string"""
        parts = str(delta).split(':')
        h, m, s = parts[0], parts[1], parts[2][:-7]
        ms = int(ceil(int(parts[2][-6:]) / 1000))
        return f"{h:02s}[h] {m:02s}[m] {s:02s}[s] {ms:02d}[ms]"

    def wrapper(*args) -> Any:
        """Execute the callable and return"""
        start = datetime.now()
        ret = func(*args)
        log.debug(f"@@@ [{func.__name__}] Time elapsed\t{timedelta_to_str(datetime.now() - start)}")
        return ret

    return wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.hspylib.core.datasource.cassandra
      @file: cassandra_repository.py
   @created: Sat, 12 Nov 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import contextlib
import logging as log
from typing import List, Optional, Set, Tuple, TypeVar

from cassandra import UnresolvableContactPoints
from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import Cluster, NoHostAvailable
from retry import retry

from hspylib.core.datasource.cassandra.cassandra_configuration import CassandraConfiguration
from hspylib.core.datasource.crud_entity import CrudEntity
from hspylib.core.datasource.db_repository import Connection, Cursor, DBRepository, ResultSet, Session
from hspylib.core.datasource.identity import Identity
from hspylib.core.exception.exceptions import DatabaseConnectionError, DatabaseError
from hspylib.core.tools.namespace import Namespace
from hspylib.core.tools.text_tools import quote

T = TypeVar('T', bound=CrudEntity)


class CassandraRepository(DBRepository[T, CassandraConfiguration]):
    """Implementation of a data access layer for a cassandra persistence store."""

    def __init__(self, config: CassandraConfiguration):
        super().__init__(config)

    @property
    def secure_bundle_path(self) -> str:
        return self._config.secure_bundle_path

    @property
    def protocol_version(self) -> int:
        return self._config.protocol_version

    @property
    def connect_timeout(self) -> int:
        return self._config.connect_timeout

    @property
    def default_timeout(self) -> int:
        return self._config.default_timeout

    @retry(tries=3, delay=2, backoff=3, max_delay=30)
    def _create_session(self) -> Tuple[Connection, Cursor]:
        log.debug(f"{self.logname} Attempt to connect to database: {str(self)}")
        auth_provider = PlainTextAuthProvider(username=self.username, password=self.password)
        if self.secure_bundle_path:
            log.info(f"{self.logname} Using Astra Security Bundle: {self.secure_bundle_path}")
            cloud_config = {'secure_connect_bundle': self.secure_bundle_path}
            cursor = Cluster(protocol_version=self.protocol_version, auth_provider=auth_provider,
                             cloud=cloud_config)
        else:
            log.info(f"{self.logname} Attempt to connect to Astra: "
                     f"{self.username}@{self.hostname}:{self.port}/{self.database}")
            cursor = Cluster(protocol_version=self.protocol_version, auth_provider=auth_provider,
                             contact_points=[self.hostname], port=self.port)
        session = cursor.connect()
        cursor.connect_timeout = self.connect_timeout
        session.default_timeout = self.default_timeout
        return cursor, session

    @contextlib.contextmanager
    def _session(self) -> Session:
        conn, dbs = None, None
        try:
            conn, dbs = self._create_session()
            log.debug(f"{self.logname} Successfully connected to database: {self.info} [ssid={hash(dbs)}]")
            yield dbs
        except (UnresolvableContactPoints, NoHostAvailable) as err:
            raise DatabaseConnectionError(f"Unable to open/execute-on database session => {err}") from err
        finally:
            if conn:
                log.debug(f"{self.logname} Shutting down connection [ssid={hash(dbs)}]")
                conn.shutdown()

    def execute(self, sql_statement: str, **kwargs) -> Tuple[int, Optional[ResultSet]]:
        """TODO"""
        with self._session() as dbs:
            try:
                rows = []
                log.debug(f"{self.logname} Executing SQL statement {sql_statement} [ssid={hash(dbs)}]:\n"
                          f"\t|-Arguments: {str([f'{k}={v}' for k, v in kwargs.items()])}\n"
                          f"\t|-Statement: {sql_statement}")
                rs = dbs.execute(sql_statement, **kwargs)
                if rs.current_rows:
                    list(map(rows.append, rs.current_rows))
                return len(rs.current_rows), rows
            except Exception as err:
                raise DatabaseError(f"Unable to execute statement => {sql_statement}") from err

    def count(self) -> int:
        sql = f"SELECT COUNT(*) FROM {self.table_name()}"
        return int(self.execute(sql)[1][0][0])

    def delete(self, entity: CrudEntity) -> None:
        self.delete_by_id(entity.identity)

    def delete_by_id(self, entity_id: Identity) -> None:
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"DELETE FROM {self.database}.{self.table_name()} WHERE " + ' AND '.join(clauses)
        self.execute(sql)

    def delete_all(self, entities: List[T]) -> None:
        values, s = [], entities[0]
        list(map(lambda e: values.append(str(e.values)), entities))
        sql = f"DELETE FROM {self.database}.{self.table_name()} " \
              f"WHERE ({s.as_columns()}) IN ({', '.join(values)}) "
        self.execute(sql)

    def save(self, entity: T) -> None:
        columns, ids = entity.as_columns(), set(entity.identity.attributes)
        sql = f"INSERT INTO {self.database}.{self.table_name()} " \
              f"({columns}) VALUES {entity.values} "
        self.execute(sql)

    def save_all(self, entities: List[T]) -> None:
        values, sample, inserts = [], entities[0], []
        columns, ids = sample.as_columns(), set(sample.identity.attributes)
        list(map(lambda e: values.append(str(e.values)), entities))
        list(map(lambda v: inserts.append(
            f"INSERT INTO {self.database}.{self.table_name()} ({columns}) VALUES {v} "), values))
        sql = f"BEGIN BATCH " \
              f"{'; '.join(inserts)} " \
              f"APPLY BATCH;"
        self.execute(sql)

    def find_all(
        self,
        fields: Optional[Set[str]] = None,
        filters: Optional[Namespace] = None,
        order_bys: Optional[List[str]] = None,
        limit: int = 500, offset: int = 0) -> List[T]:

        fields = '*' if not fields else ', '.join(fields)
        clauses = list(filter(None, [f for f in filters.values])) if filters else None
        orders = list(filter(None, order_bys)) if order_bys and clauses else None  # Order by require filters
        sql = f"SELECT {fields} FROM {self.database}.{self.table_name()} " \
              f"{('WHERE ' + ' AND '.join(clauses)) if clauses else ''} " \
              f"{('ORDER BY ' + ', '.join(orders)) if orders else ''} " \
              f"LIMIT {limit}"

        return list(map(self.to_entity_type, self.execute(sql)[1]))

    def find_by_id(
        self,
        entity_id: Identity,
        fields: Optional[Set[str]] = None) -> Optional[T]:

        fields = '*' if not fields else ', '.join(fields)
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"SELECT {fields} FROM {self.database}.{self.table_name()} " \
              f"WHERE {' AND '.join(clauses)}"
        result = next((e for e in self.execute(sql)[1]), None)

        return self.to_entity_type(result) if result else None

    def exists_by_id(self, entity_id: Identity) -> bool:
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"SELECT EXISTS(SELECT 1 FROM {self.table_name()} WHERE {' AND '.join(clauses)})"

        return self.execute(sql)[1][0][0] > 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: hspylib.core.tools
      @file: ttl_keyring_be.py
   @created: Tue, 4 Oct 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import json
from datetime import datetime
from typing import Optional

import keyring
from keyring.backends.chainer import ChainerBackend
from keyring.errors import PasswordDeleteError, PasswordSetError

from hspylib.core.preconditions import check_not_none
from hspylib.core.tools.zoned_datetime import now_ms


class TTLKeyringBE(ChainerBackend):

    priority = 1

    def __init__(self, ttl_minutes: int = 15, ttl_seconds=0) -> None:
        super().__init__()
        self._ttl_minutes = ttl_minutes
        self._ttl_seconds = ttl_seconds

    def get_pws(self, service_name: str, username: str) -> str:
        return super().get_password(service_name, username)

    def set_password(self, service_name: str, username: str, password: str) -> None:
        """Set password for the username of the service."""
        try:
            check_not_none(service_name, username, password)
            expires_sec = now_ms() + ((self._ttl_minutes * 60) + self._ttl_seconds)
            passwd_obj = {'sn': service_name, 'un': username, 'pw': password, 'ttl': expires_sec}
            super().set_password(service_name, username, json.dumps(passwd_obj))
        except PasswordSetError:
            pass  # it does not matter if the password set failed.

    def get_password(self, service_name: str, username: str) -> Optional[str]:
        """Get password of the username for the service."""
        check_not_none(service_name, username)
        if passwd_str := self.get_pws(service_name, username):
            passwd_obj = json.loads(passwd_str)
            dt_object = datetime.fromtimestamp(passwd_obj['ttl'])
            if now_ms() - dt_object.timestamp() > 0:
                keyring.delete_password(service_name, username)
            else:
                return passwd_obj['pw']

        return None

    def delete_password(self, service_name: str, username: str) -> None:
        """Delete the password for the username of the service."""
        check_not_none(service_name, username)
        try:
            super().delete_password(service_name, username)
        except PasswordDeleteError:
            pass  # it does not matter if the password does not exist.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.vt100
      @file: vt_codes.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import re
from enum import auto
from typing import Callable, Optional

from hspylib.core.enums.enumeration import Enumeration
from hspylib.modules.cli.vt100.vt_100 import Vt100


def vt_print(vt100_str: str, end: str = '') -> None:
    """Print a vt-100 encoded string. VT-100 string will contain one or more %VT-100-CODE% """
    print(VtCodes.decode(vt100_str), end=end)


class VtCodes(Enumeration):
    """VT-100 escape codes"""

    # @formatter:off
    CSV = Vt100.save_cursor()           # ^[7 -> Save cursor position and attributes
    CRE = Vt100.restore_cursor()        # ^[8 -> Restore cursor position and attributes
    RIS = Vt100.reset()                 # ^[c -> Reset terminal to initial state

    SAW = Vt100.set_auto_wrap(True)     # ^[?7h  -> Set auto-wrap mode
    UAW = Vt100.set_auto_wrap(False)    # ^[?7l  -> Unset auto-wrap mode
    SSC = Vt100.set_show_cursor(True)   # ^[?25h -> Set show cursor
    USC = Vt100.set_show_cursor(False)  # ^[?25l -> Unset show cursor

    ED0 = Vt100.clear_screen()          # ^[[J  -> Clear screen from cursor down
    ED1 = Vt100.clear_screen(1)         # ^[[1J -> Clear screen from cursor up
    ED2 = Vt100.clear_screen(2)         # ^[[2J -> Clear entire screen

    EL0 = Vt100.clear_line()            # ^[[K  -> Clear line from cursor right
    EL1 = Vt100.clear_line(1)           # ^[[1K -> Clear line from cursor left
    EL2 = Vt100.clear_line(2)           # ^[[2K -> Clear entire line

    HOM = Vt100.cursor_pos()            # ^[[H  -> Move cursor to upper left corner

    # The following entries must defined as auto(), so they can be invoked as Callable

    MOD = auto()                        # ^[[<m1;m2;m3>m  -> Set terminal modes
    CUP = auto()                        # ^[[<v>;<h>H     -> Move cursor to screen location <v,h>
    CUU = auto()                        # ^[[<n>A         -> Move cursor up n lines
    CUD = auto()                        # ^[[<n>B         -> Move cursor down n lines
    CUF = auto()                        # ^[[<n>C         -> Move cursor right n lines
    CUB = auto()                        # ^[[<n>D         -> Move cursor left n lines
    # @formatter:on

    # For all mnemonics that take arguments we need to include in this map, so we can call it

    @staticmethod
    def _vt100_fnc(command: str) -> Optional[Callable]:
        """TODO"""
        fnc = None
        match command:
            case "MOD":
                fnc = Vt100.mode
            case "CUP":
                fnc = Vt100.cursor_pos
            case "CUU":
                fnc = Vt100.cursor_move_up
            case "CUD":
                fnc = Vt100.cursor_move_down
            case "CUF":
                fnc = Vt100.cursor_move_forward
            case "CUB":
                fnc = Vt100.cursor_move_backward
        return fnc

    @classmethod
    def decode(cls, input_string: str) -> str:
        """Decode the string into a VT_CODE enum"""
        results = re.findall(r'%([a-zA-Z0-9]+)(\([0-9]+(;[0-9]+)*\))?%', input_string)
        for nextResult in results:
            mnemonic = nextResult[0]
            if mnemonic in VtCodes.names():
                args = nextResult[1][1:-1] if nextResult[1] else ''
                # Command has args, so, we need to call the vt100 function
                if args:
                    fnc = VtCodes.value_of(mnemonic).__call__
                    input_string = input_string.replace(
                        f'%{mnemonic + nextResult[1]}%', fnc(args) if fnc else '')
                else:
                    input_string = input_string.replace(
                        f'%{mnemonic}%', str(VtCodes.value_of(mnemonic).value))

        return input_string

    def __call__(self, *args, **kwargs) -> str:
        return VtCodes._vt100_fnc(self.name)(args[0])

    def __str__(self) -> str:
        return str(self.value)

    def code(self) -> str:
        """Decode the string into a VT_CODE enum"""
        return str(self)

    def placeholder(self) -> str:
        """Decode the string into a VT_CODE enum"""
        return f"%{self.name}%"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.tools
      @file: commons.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import inspect
import logging as log
import os
import pathlib
import sys
from datetime import timedelta
from typing import Optional, Set, Tuple, Type
from urllib.parse import urlparse

from hspylib.core.constants import TRUE_VALUES
from hspylib.core.enums.charset import Charset
from hspylib.core.tools.validator import Validator
from hspylib.modules.cli.vt100.vt_codes import VtCodes
from hspylib.modules.cli.vt100.vt_colors import VtColors

# pylint: disable=consider-using-f-string
FILE_LOG_FMT = '{}\t{} {} {} {} {} '.format(
    '%(asctime)s',
    '%(levelname)-5.5s',
    '%(filename)s::',
    '%(message)s',
    '%(funcName)s(@Line:%(lineno)d) -',
    '%(threadName)-12.12s',
)

CONSOLE_LOG_FMT = '{}\t{} {} {} {} '.format(
    '%(levelname)-5.5s',
    '%(message)s',
    '%(funcName)s(@Line:%(lineno)d)',
    '%(asctime)s -',
    '%(threadName)-12.12s'
)


def log_init(
    filename: Optional[str] = None,
    filemode: str = 'a',
    level: int = log.DEBUG,
    log_format: str = FILE_LOG_FMT,
    clear_handlers: bool = True,
    console_enable: bool = False,
    file_enable: bool = True) -> bool:
    """Initialize the system logger"""

    if file_enable and not os.path.exists(filename):
        touch_file(filename)

    # if someone tried to log something before log_init is called, Python creates a default handler that is going to
    # mess our logs. Remove handlers if there is any.
    root, handlers = log.getLogger(), set()

    if clear_handlers:
        if root.handlers:
            for handler in root.handlers:
                handler.close()
                root.removeHandler(handler)

    if file_enable:
        file_formatter = log.Formatter(log_format)
        file_handler = log.FileHandler(filename=filename, mode=filemode)
        file_handler.setFormatter(file_formatter)
        handlers.add(file_handler)

    if console_enable:
        console_formatter = log.Formatter(CONSOLE_LOG_FMT, "%Y-%m-%d %H:%M:%S")
        console_handler = log.StreamHandler(sys.stdout)
        console_handler.setFormatter(console_formatter)
        handlers.add(console_handler)

    log.basicConfig(level=level, handlers=handlers)

    return (os.path.exists(filename or '') and file_enable) or console_enable


def is_debugging() -> bool:
    """Whether the program is running under debug mode."""
    for frame in inspect.stack():
        if frame[1].endswith("pydevd.py"):
            return True
    return False


def dirname(filepath: str) -> str:
    """Retrieve the directory of the specified filepath"""
    return os.path.dirname(os.path.realpath(filepath))


def run_dir() -> str:
    """Retrieve the application's root directory"""
    return sys.path[0]


def get_path(filepath: str) -> pathlib.Path:
    """Retrieve the Path of the file"""
    return pathlib.Path(filepath).parent


def sysout(string: str, end: str = '\n') -> None:
    """Print the unicode input_string decoding vt100 placeholders
    :param string: values to be printed to sys.stdout
    :param end: string appended after the last value, default a newline
    """
    if Validator.is_not_blank(string):
        msg = VtColors.colorize(VtCodes.decode(f"{string}%NC%"))
        print(msg, file=sys.stdout, flush=True, end=end)


def syserr(string: str, end: str = '\n') -> None:
    """Print the unicode input_string decoding vt100 placeholders
    :param string: values to be printed to sys.stderr
    :param end: string appended after the last value, default a newline
    """
    if Validator.is_not_blank(string):
        msg = VtColors.colorize(VtCodes.decode(f"%RED%{string}%NC%"))
        print(msg, file=sys.stderr, flush=True, end=end)


def class_attribute_names(clazz: Type) -> Optional[Tuple]:
    """Retrieve all attribute names of the class
    :param clazz: The class to retrieve the attribute names
    """
    return tuple(vars(clazz()).keys()) if clazz else None


def class_attribute_values(instance: dict) -> Optional[Tuple]:
    """Retrieve all attribute values of the class
    :param instance: The class to retrieve the attribute values
    """
    return tuple(instance.values()) if instance else None


def str_to_bool(string: str, true_values: Set[str] = None) -> bool:
    """Convert a string to boolean
    :param string: The string to be converted
    :param true_values: The list of strings that will become True value
    """
    return string is not None and string.lower() in (true_values or TRUE_VALUES)


def safe_del_file(filename: str, on_not_found_except: bool = False) -> bool:
    """Delete the file specified by filename. If the file is not found, raises an exception if on_not_found_except is
    True; otherwise return False.
    :param filename: the name of the file to be checked
    :param on_not_found_except: boolean parameter to raise an exception if the file is not found.
    """
    if os.path.exists(filename):
        os.remove(filename)
    else:
        if on_not_found_except:
            raise FileNotFoundError(f'File was not found on the system: {filename}')
        else:
            return False

    return True


def file_is_not_empty(filename: str) -> bool:
    """Check whether the file is empty or not.
    :param filename: the name of the file to be modified. If the file does not exist; return False
    """
    return os.path.exists(filename) and os.stat(filename).st_size > 0


def touch_file(filename: str, encoding: str = Charset.UTF_8.value) -> None:
    """Change file modification time
    :param filename: the name of the file to be modified
    :param encoding: the file encoding
    """
    with open(filename, 'a', encoding=encoding):
        os.utime(filename, None)


def human_readable_bytes(size_in_bytes: int) -> Tuple[str, str]:
    """Return a Human readable bytes and unit
    :param size_in_bytes: the size to be formatted
    """

    byte_size = float(size_in_bytes)
    kb, mb, gb, tb = 2 ** 10, 2 ** 20, 2 ** 30, 2 ** 40

    if 0 <= byte_size <= kb:
        ret_val = f'{byte_size:3.2f}'
        ret_unit = '[B]'
    elif kb < byte_size <= mb:
        ret_val = f'{byte_size / kb:3.2f}'
        ret_unit = '[Kb]'
    elif mb < byte_size <= gb:
        ret_val = f'{byte_size / mb:3.2f}'
        ret_unit = '[Mb]'
    elif gb < byte_size <= tb:
        ret_val = f'{byte_size / gb:3.2f}'
        ret_unit = '[Gb]'
    else:
        ret_val = f'{byte_size / tb:3.2f}'
        ret_unit = '[Tb]'

    return ret_val, ret_unit


def human_readable_time(time_microseconds: int) -> str:
    """Return a Human readable formatted time
    :param time_microseconds: the time to be formatted
    """
    delta = timedelta(microseconds=time_microseconds)
    total_seconds = delta.seconds
    seconds = total_seconds % 60
    minutes = total_seconds / 60 % 60
    hours = total_seconds / 3600
    microseconds = delta.microseconds
    # Using format: HH:MM:SS.uuuuuu
    return f"{hours:02d}:{minutes:02d}:{seconds:02d}.{microseconds:06d}"


def build_url(base_url: str, scheme: str = 'http') -> str:
    """TODO"""
    p_url = urlparse(base_url)
    if not p_url.scheme:
        base_url = p_url._replace(scheme=scheme).geturl()

    return base_url



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.hspylib.core.datasource
      @file: crud_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import json
from abc import abstractmethod
from typing import Generic, List, Optional, Set, TypeVar

from hspylib.core.datasource.crud_entity import CrudEntity
from hspylib.core.datasource.identity import Identity
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.tools.namespace import Namespace

T = TypeVar('T', bound=CrudEntity)


class CrudRepository(Generic[T], metaclass=Singleton):
    """Interface for generic CRUD operations on a repository for a specific type."""

    @property
    def logname(self) -> str:
        """TODO"""
        return self.__class__.__name__.split('_')[0]

    @abstractmethod
    def count(self) -> int:
        """TODO"""

    @abstractmethod
    def delete(self, entity: T) -> None:
        """TODO"""

    @abstractmethod
    def delete_by_id(self, entity_id: Identity) -> None:
        """TODO"""

    @abstractmethod
    def delete_all(self, entities: List[T]) -> None:
        """TODO"""

    @abstractmethod
    def save(self, entity: T) -> None:
        """TODO"""

    @abstractmethod
    def save_all(self, entities: List[T]) -> None:
        """TODO"""

    @abstractmethod
    def find_all(
        self,
        fields: Optional[Set[str]] = None,
        filters: Optional[Namespace] = None,
        order_bys: Optional[List[str]] = None,
        limit: int = 500, offset: int = 0) -> List[T]:
        """TODO"""

    @abstractmethod
    def find_by_id(
        self,
        entity_id: Identity,
        fields: Optional[Set[str]] = None) -> Optional[T]:
        """TODO"""

    def exists_by_id(self, entity_id: Identity) -> bool:
        """TODO"""

    @abstractmethod
    def to_entity_type(self, entity_dict: dict | tuple) -> T:
        """TODO"""

    def to_entity_list(self, json_string: str, filters: Optional[Namespace] = None) -> List[T]:
        """Return filtered entries from the json_string as a list"""
        ret_list = []
        for value in json.loads(json_string).values():
            if not filters or all(k in value and value[k] == v for k, v in filters.items()):
                ret_list.append(self.to_entity_type(value))
        return ret_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.hspylib.core.datasource.mysql
      @file: mysql_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import contextlib
import logging as log
from typing import List, Optional, Set, Tuple, TypeVar

import pymysql
from pymysql import Connection, Error
from pymysql.cursors import Cursor
from retry import retry

from hspylib.core.datasource.crud_entity import CrudEntity
from hspylib.core.datasource.db_configuration import DBConfiguration
from hspylib.core.datasource.db_repository import DBRepository, ResultSet, Session
from hspylib.core.datasource.identity import Identity
from hspylib.core.exception.exceptions import DatabaseConnectionError, DatabaseError
from hspylib.core.tools.namespace import Namespace
from hspylib.core.tools.text_tools import quote

T = TypeVar('T', bound=CrudEntity)


class MySqlRepository(DBRepository[T, DBConfiguration]):
    """Implementation of a data access layer for a MySql persistence store."""

    def __init__(self, config: DBConfiguration):
        super().__init__(config)

    @retry(tries=3, delay=2, backoff=3, max_delay=30)
    def _create_session(self) -> Tuple[Connection, Cursor]:
        log.debug(f"{self.logname} Attempt to connect to database: {str(self)}")
        conn = pymysql.connect(
            host=self.hostname,
            user=self.username,
            port=self.port,
            password=self.password,
            database=self.database)
        return conn, conn.cursor()

    @contextlib.contextmanager
    def _session(self) -> Session:
        conn, dbs = None, None
        try:
            conn, dbs = self._create_session()
            log.debug(f"{self.logname} Successfully connected to database: {self.info} [ssid={hash(dbs)}]")
            yield dbs
        except Error as err:
            raise DatabaseConnectionError(f"Unable to open/execute-on database session => {err}") from err
        finally:
            if conn:
                log.debug(f"{self.logname} Closing connection [ssid={hash(dbs)}]")
                conn.commit()
                conn.close()

    def execute(self, sql_statement: str, **kwargs) -> Tuple[int, Optional[ResultSet]]:
        """TODO"""
        with self._session() as dbs:
            try:
                rows = []
                log.debug(f"{self.logname} Executing SQL statement {sql_statement} [ssid={hash(dbs)}]:\n"
                          f"\t|-Arguments: {str([f'{k}={v}' for k, v in kwargs.items()])}\n"
                          f"\t|-Statement: {sql_statement}")
                if (rows_affected := dbs.execute(sql_statement, **kwargs) or 0) > 0:
                    list(map(rows.append, dbs))
                return rows_affected, rows
            except (pymysql.err.ProgrammingError, pymysql.err.OperationalError) as err:
                raise DatabaseError(f"Unable to execute statement => {sql_statement}") from err

    def count(self) -> int:
        sql = f"SELECT COUNT(*) FROM {self.table_name()}"
        return int(self.execute(sql)[1][0][0])

    def delete(self, entity: CrudEntity) -> None:
        self.delete_by_id(entity.identity)

    def delete_by_id(self, entity_id: Identity) -> None:
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"DELETE FROM {self.table_name()} WHERE " + ' AND '.join(clauses)
        self.execute(sql)

    def delete_all(self, entities: List[T]) -> None:
        values, s = [], entities[0]
        list(map(lambda e: values.append(str(e.values)), entities))
        sql = f"DELETE FROM " \
              f"{self.table_name()} WHERE ({s.as_columns()}) IN ({', '.join(values)}) "
        self.execute(sql)

    def save(self, entity: T) -> None:
        columns, ids = entity.as_columns(), set(entity.identity.attributes)
        sql = f"INSERT INTO " \
              f"{self.table_name()} ({columns}) VALUES {entity.values} AS new " \
              f"ON DUPLICATE KEY UPDATE {entity.as_column_set(prefix='new.', exclude=ids)}"
        self.execute(sql)

    def save_all(self, entities: List[T]) -> None:
        values, sample = [], entities[0]
        columns, ids = sample.as_columns(), set(sample.identity.attributes)
        list(map(lambda e: values.append(str(e.values)), entities))
        sql = f"INSERT INTO " \
              f"{self.table_name()} ({columns}) VALUES {', '.join(values)} AS new " \
              f"ON DUPLICATE KEY UPDATE {sample.as_column_set(prefix='new.', exclude=ids)}"
        self.execute(sql)

    def find_all(
        self,
        fields: Optional[Set[str]] = None,
        filters: Optional[Namespace] = None,
        order_bys: Optional[List[str]] = None,
        limit: int = 500, offset: int = 0) -> List[T]:

        fields = '*' if not fields else ', '.join(fields)
        clauses = list(filter(None, [f for f in filters.values])) if filters else None
        orders = list(filter(None, order_bys)) if order_bys else None
        sql = f"SELECT {fields} FROM {self.table_name()} " \
              f"{('WHERE ' + ' AND '.join(clauses)) if clauses else ''} " \
              f"{('ORDER BY ' + ', '.join(orders)) if orders else ''} " \
              f"LIMIT {limit} OFFSET {offset}"

        return list(map(self.to_entity_type, self.execute(sql)[1]))

    def find_by_id(
        self,
        entity_id: Identity,
        fields: Optional[Set[str]] = None) -> Optional[T]:

        fields = '*' if not fields else ', '.join(fields)
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"SELECT {fields} FROM {self.table_name()} " \
              f"WHERE {' AND '.join(clauses)}"
        result = next((e for e in self.execute(sql)[1]), None)

        return self.to_entity_type(result) if result else None

    def exists_by_id(self, entity_id: Identity) -> bool:
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"SELECT EXISTS(SELECT 1 FROM {self.table_name()} WHERE {' AND '.join(clauses)})"

        return self.execute(sql)[1][0][0] > 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.qt
      @file: stream_capturer.py
   @created: Wed, 30 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import io
import logging as log
from contextlib import redirect_stderr, redirect_stdout
from time import sleep

from PyQt5.QtCore import pyqtSignal, QThread

from hspylib.core.preconditions import check_argument
from hspylib.core.tools.commons import is_debugging, syserr


class StreamCapturer(QThread):
    """QThread to captures stdout and/or stderr messages and send them via PyQt Signal"""

    stdoutCaptured = pyqtSignal(str)
    stderrCaptured = pyqtSignal(str)

    class StdoutWorker(QThread):
        """QThread worker to capture stdout messages"""

        streamCaptured = pyqtSignal(str)

        def __init__(self, parent: 'StreamCapturer', poll_interval: float):
            super().__init__()
            self._poll_interval = poll_interval
            self._parent = parent

        def run(self):
            self.setObjectName(f"stdout-worker-{hash(self)}")
            with io.StringIO() as buf, redirect_stdout(buf):
                while not self._parent.isFinished():
                    output = buf.getvalue()
                    if output and output != '':
                        log.info(output)
                        self.streamCaptured.emit(output)
                        buf.truncate(0)
                    sleep(self._poll_interval)

    class StderrWorker(QThread):
        """QThread worker to capture stderr messages"""

        streamCaptured = pyqtSignal(str)

        def __init__(self, parent: 'StreamCapturer', poll_interval: float):
            super().__init__()
            self._poll_interval = poll_interval
            self._parent = parent

        def run(self):
            self.setObjectName('stderr-worker')
            with io.StringIO() as buf, redirect_stderr(buf):
                while not self._parent.isFinished():
                    output = buf.getvalue()
                    if output and output != '':
                        log.error(output)
                        self.streamCaptured.emit(output)
                        buf.truncate(0)
                    sleep(self._poll_interval)

    def __init__(
        self,
        capture_stderr: bool = True,
        capture_stdout: bool = True,
        stdout_poll_interval: float = 0.5,
        stderr_poll_interval: float = 0.5):

        check_argument(capture_stderr or capture_stdout, 'At least one capturer must be started')
        super().__init__()
        self.setObjectName('stream-capturer')
        self._capture_stdout = capture_stdout
        self._capture_stderr = capture_stderr
        self._poll_interval = stdout_poll_interval + stderr_poll_interval

        if self._capture_stderr:
            self._stderr_capturer = self.StderrWorker(self, stderr_poll_interval)
            self._stderr_capturer.streamCaptured.connect(self.stderrCaptured.emit)
        if self._capture_stdout:
            self._stdout_capturer = self.StdoutWorker(self, stdout_poll_interval)
            self._stdout_capturer.streamCaptured.connect(self.stdoutCaptured.emit)

    def run(self) -> None:
        if self._capture_stderr:
            self._stderr_capturer.start()
        if self._capture_stdout:
            self._stdout_capturer.start()
        while not self.isFinished():
            sleep(self._poll_interval)

    def start(self, priority: QThread.Priority = QThread.NormalPriority) -> None:
        if not is_debugging():
            super().start(priority)
        else:
            syserr('Stderr/Stdout capture is not started in debugging mode')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.qt.views
      @file: qt_view.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import os
from abc import ABC
from typing import Optional, Tuple, Type

from PyQt5 import uic
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QWidget

from hspylib.core.preconditions import check_argument, check_state
from hspylib.core.tools.commons import run_dir


class QtView(ABC):
    """TODO"""

    MAIN_QT_VIEW_UI = 'main_qt_view.ui'

    @staticmethod
    def load_form(
        form_file: str,
        load_dir: str) -> Tuple[Type, Type]:
        """Load the ui form from the .ui file"""

        form_dir = load_dir if load_dir else f"{run_dir()}/resources/forms/"
        check_argument(
            os.path.exists(form_dir) and os.path.isdir(form_dir),
            "Load dir {} does not exist or is not a folder", form_dir)
        filepath = f"{form_dir}/{form_file}"
        check_state(
            os.path.exists(filepath) and os.path.isfile(filepath) and filepath.lower().endswith('.ui'),
            "Form file {} does not exist or it is not a valid UI form file", form_file)

        return uic.loadUiType(filepath)

    def __init__(self, ui_file: str = MAIN_QT_VIEW_UI, load_dir: str = None, parent: Optional[QWidget] = None):
        ui_clazz, window_clazz = self.load_form(ui_file, load_dir)
        # Must come after the initialization above
        check_state(ui_clazz is not None and window_clazz is not None,
                    'Unable to initialize UI and Window objects')
        self.window, self.ui = window_clazz(), ui_clazz()
        self.ui.setupUi(self.window)
        self.parent = parent

    def show(self) -> None:
        """Show the main widget"""
        self.window.show()

    def set_default_font(self, default_font: QFont = QFont('Courier New', 14)):
        """Set font for all UI components at once"""
        widgets = list(filter(lambda o: hasattr(getattr(self.ui, o), 'setFont'), vars(self.ui)))
        list(map(lambda w: getattr(self.ui, w).setFont(default_font), widgets))



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.123
#
# Package: main.hspylib.modules.qt
"""Package initialization."""

__all__ = [
  'promotions', 
  'qt_application', 
  'stream_capturer', 
  'views'
]
__version__ = '0.11.123'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   main.addons.appman
      @file: appman.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import os
from textwrap import dedent
from typing import List

import urllib3

from hspylib.addons.appman.appman_enums import AppType, Extension
from hspylib.core.enums.exit_status import ExitStatus
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.preconditions import check_argument
from hspylib.core.tools.commons import get_path, syserr, sysout
from hspylib.core.tools.namespace import Namespace
from hspylib.core.tools.text_tools import camelcase
from hspylib.modules.cli.application.application import Application
from hspylib.modules.cli.tui.extra.minput.input_validator import InputValidator
from hspylib.modules.cli.tui.extra.minput.minput import MenuInput, minput
from hspylib.modules.cli.vt100.terminal import Terminal
from hspylib.modules.fetch.fetch import get

HERE = get_path(__file__)

INITIAL_REVISION = '0.1.0'

WELCOME_MESSAGE = f'My Application v{INITIAL_REVISION}'

# Disable this warning because we trust our project repo
urllib3.disable_warnings()


class AppManager(metaclass=Singleton):
    """HSPyLib application manager that helps creating HSPyLib based applications and widgets"""

    # The directory containing all template files
    TEMPLATES = (HERE / "templates")

    # The general gradle properties
    GRADLE_PROPS = dedent("""
        app_name    = '{}'
        app_version = '{}'

        pythonVersion=3
        pyrccVersion=5

        author="<Author>")
        mailTo="<MailTo")
        siteUrl="<SiteUrl>")
    """)

    @staticmethod
    def prompt() -> Namespace:
        """When no input is provided, prompt the user for the info. """
        # @formatter:off
        form_fields = MenuInput.builder() \
            .field() \
                .label('App Name') \
                .validator(InputValidator.words()) \
                .min_max_length(1, 40) \
                .value('my_app') \
                .build() \
            .field() \
                .label('App Type') \
                .itype('select') \
                .value(f"{AppType.APP}|{AppType.QT_APP}|{AppType.WIDGET}") \
                .build() \
            .field() \
                .label('Dest Dir') \
                .validator(InputValidator.anything()) \
                .min_max_length(3, 80) \
                .value(os.getenv('HOME', os.getcwd())) \
                .build() \
            .field() \
                .label('Initialize gradle') \
                .itype('checkbox') \
                .value(True) \
                .build() \
            .field() \
                .label('Initialize git') \
                .itype('checkbox') \
                .value(True) \
                .build() \
            .build()
        # @formatter:on
        return minput(form_fields)

    def __init__(self, parent_app: Application):
        self._parent_app = parent_app
        self._app_name = None
        self._app_dir = None
        self._init_gradle_flag = False
        self._init_git_flag = False

    def create(self, app_name: str, app_type: AppType, app_ext: List[Extension], dest_dir: str) -> None:
        """Create the application based on the parameters"""
        sysout(f'Creating "{app_name}" at {dest_dir}')
        try:
            check_argument(os.path.exists(dest_dir), 'Destination not found: {}', dest_dir)
            self._app_name = app_name
            if app_type == AppType.APP:
                self._app_dir = f'{dest_dir}/{app_name}'
                self._create_app(app_name, app_ext)
            elif app_type == AppType.WIDGET:
                self._app_dir = f'{dest_dir}'
                self._create_widget(app_name)
            elif app_type == AppType.QT_APP:
                self._app_dir = f'{dest_dir}/{app_name}'
                self._create_qt_app(app_name, app_ext)
            else:
                raise TypeError(f'Unsupported application type: {app_type}')
        except OSError as err:
            syserr(f'Could not create application "{app_name}"!\n  => {str(err)}')
            self._parent_app.exit(ExitStatus.FAILED.value)

        sysout(f"Successfully created the {app_type.value} {app_name}")

    def _create_app(self, app_name: str, extensions: List[Extension]) -> None:
        """Create a Simple HSPyLib application"""
        sysout(f'Application: {app_name}')
        self._create_base_app_struct(app_name)
        self._mkfile(
            f'src/main/{self._app_name}/__main__.py',
            (self.TEMPLATES / "main.py.tpl").read_text().replace('%APP_NAME%', self._app_name))
        self._apply_extensions(extensions, app_name)

    def _create_qt_app(self, app_name: str, extensions: List[Extension]) -> None:
        """Create an HSPyLib QT application"""
        sysout(f'QT Application: {app_name}')
        self._create_base_app_struct(app_name)
        self._mkdir(f'src/main/{self._app_name}/resources/forms')
        self._mkfile(
            f'src/main/{self._app_name}/resources/forms/main_qt_view.ui',
            (self.TEMPLATES / "main_qt_view.ui.tpl").read_text())
        self._mkdir(f'src/main/{self._app_name}/view')
        self._mkfile(
            f'src/main/{self._app_name}/view/main_qt_view.py',
            (self.TEMPLATES / "main_qt_view.py.tpl").read_text().replace('%APP_NAME%', self._app_name))
        self._mkfile(
            f'src/main/{self._app_name}/__main__.py',
            (self.TEMPLATES / "main_qt.py.tpl").read_text().replace('%APP_NAME%', self._app_name))
        self._apply_extensions(extensions, app_name)

    def _create_widget(self, app_name: str) -> None:
        """Create an HSPyLib Widget application"""
        widget_name = camelcase(app_name).replace('_', '').replace(' ', '')
        sysout(f'Widget: {widget_name}')
        self._mkfile(
            f"widget_{app_name.lower()}.py",
            (self.TEMPLATES / "widget.py.tpl").read_text().replace('_WIDGET_NAME_', f"{widget_name}")
        )

    def _create_base_app_struct(self, app_name: str) -> None:
        """Create HSPyLib application structure"""
        self._mkdir()
        self._mkfile('README.md', f'# {app_name}')
        self._mkfile('MANIFEST.in')
        self._mkdir('src')
        self._mkdir('src/test')
        self._mkfile('src/test/test_main.py', (self.TEMPLATES / "test_main.py.tpl").read_text())
        self._mkdir('src/test/resources')
        self._mkfile('src/test/resources/application-test.properties', '# Main test application property file')
        self._mkdir(f'src/main')
        self._mkdir(f'src/main/{self._app_name}')
        self._mkfile(f'src/main/{self._app_name}/__classpath__.py', (self.TEMPLATES / "classpath.py.tpl").read_text())
        self._mkfile(f'src/main/{self._app_name}/.version', INITIAL_REVISION)
        self._mkfile(f'src/main/{self._app_name}/welcome.txt', WELCOME_MESSAGE)
        self._mkdir(f'src/main/{self._app_name}/resources')
        self._mkfile(f'src/main/{self._app_name}/resources/application.properties', '# Main application property file')
        self._mkfile('.env', '# Type in here the environment variables your app requires')
        self._mkfile('run.sh', (self.TEMPLATES / "run.sh.tpl").read_text().replace('%APP_NAME%', self._app_name))

    def _apply_extensions(self, extensions: List[Extension], app_name: str):
        """Apply the selected extensions to the app"""
        result = True
        if Extension.GRADLE in extensions:
            sysout('Applying gradle extensions')
            result = self._init_gradle(app_name)
        if result and Extension.GIT in extensions:
            sysout('Initializing git repository')
            self._init_git()

    def _mkdir(self, dirname: str = '') -> None:
        """Create a directory from the destination path, or the destination path itself"""
        dir_path = f"{self._app_dir}/{dirname}"
        sysout(f'  |- {dir_path}')
        os.mkdir(dir_path)

    def _mkfile(self, filename: str, contents: str = '') -> None:
        """Create a file from the destination path with the specified contents"""
        file_path = f"{self._app_dir}/{filename}"
        sysout(f'  |- {file_path}')
        with open(f'{file_path}', 'w', encoding='utf-8') as fh:
            fh.write(contents)

    def _download_ext(self, extension: str) -> None:
        """Download a gradle extension from the HSPyLib repository"""
        resp = get(f'https://raw.githubusercontent.com/yorevs/hspylib/master/gradle/{extension}')
        check_argument(resp.status_code == HttpCode.OK, 'Unable to download {}', extension)
        self._mkfile(f'gradle/{extension}', resp.body)

    def _init_gradle(self, app_name: str) -> bool:
        """Initialize the as a gradle project"""
        sysout('Initializing gradle project [press enter to continue] ...')
        output, exit_code = Terminal.shell_exec(
            f"gradle init --project-name {app_name} --type basic --dsl groovy", cwd=self._app_dir)
        sysout(f'Gradle execution result: {exit_code}')
        sysout(output)

        if exit_code == ExitStatus.SUCCESS:
            sysout('Downloading gradle extensions ...')
            self._download_ext('badges.gradle')
            self._download_ext('build-info.gradle')
            self._download_ext('dependencies.gradle')
            self._download_ext('docgen.gradle')
            self._download_ext('docker.gradle')
            self._download_ext('pypi-publish.gradle')
            self._download_ext('python.gradle')
            sysout('Creating gradle files')
            self._mkfile(
                'gradle.properties',
                self.GRADLE_PROPS.format(self._app_name, INITIAL_REVISION).strip())
            self._mkfile(
                'build.gradle',
                (self.TEMPLATES / "build.gradle.tpl").read_text().replace('%APP_NAME%', self._app_name))
            self._mkfile('dependencies.hspd', (self.TEMPLATES / "dependencies.hspd.tpl").read_text())
            sysout('Building gradle project, please wait ...')
            output, exit_code = Terminal.shell_exec('./gradlew buildOnly', cwd=self._app_dir)
            sysout(f'Gradle execution result: {output}')

        return exit_code == ExitStatus.SUCCESS

    def _init_git(self) -> bool:
        """Initialize a git repository for the project"""
        self._mkfile('.gitignore', (self.TEMPLATES / "gitignore.tpl").read_text())
        sysout('Initializing git repository')
        output, exit_code = Terminal.shell_exec(
            'git init', cwd=self._app_dir)
        sysout(f'Git init result: {exit_code}')
        sysout(output)
        if exit_code:
            sysout('Creating first commit')
            Terminal.shell_exec(
                'git add .', cwd=self._app_dir)
            output, exit_code = Terminal.shell_exec(
                'git commit -m "First commit [@HSPyLib]"', cwd=self._app_dir)
            sysout(f'Git commit result: {exit_code}')
            sysout(output)

        return exit_code == ExitStatus.SUCCESS



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.main.hspylib.modules.qt.promotions
      @file: hlistwidget.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from typing import List, Optional, Union

from PyQt5.QtCore import pyqtSignal, QModelIndex, Qt
from PyQt5.QtGui import QCursor, QKeyEvent
from PyQt5.QtWidgets import QListWidget, QListWidgetItem, QMenu, QWidget

from hspylib.core.preconditions import check_argument, check_not_none, check_state


class HListWidget(QListWidget):
    """TODO"""

    keyPressed = pyqtSignal(int)

    @staticmethod
    def unset_flag(flags: int, flag: int) -> int:
        """TODO"""
        return flags ^ flag if flags & flag == flag else flags

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self._editable = False
        self._selectable = True
        self._items = []
        self._custom_menu_actions = []
        self._context_menu_enable = True
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._context_menu)
        self.itemChanged.connect(self.item_changed)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """Handles and forwards the key press event in the list. """
        super().keyPressEvent(event)
        self.keyPressed.emit(event.key())

    def item_changed(self, item: QListWidgetItem):
        """Handles itemChanged event to avoid adding duplicates to the list. """
        existing = self.findItems(item.text(), Qt.MatchFixedString)
        if len(existing) > 1:
            self.del_item(item)

    def addItem(self, item: QListWidgetItem) -> None:
        """Adds the specified element to this list regardless if it is present or not. """
        prev = self.item(self.count() - 1)
        if prev:
            item.setFlags(prev.flags())
        super().addItem(item)
        self._items.append(item)

    def is_empty(self) -> bool:
        """Returns true if this list contains no elements. """
        return len(self._items) == 0

    def set_item(self, item: Union[str, QListWidgetItem], flags: Union[Qt.ItemFlags, Qt.ItemFlag] = None) -> None:
        """Adds the specified element to this list if it is not already present."""
        if not self.findItems(item, Qt.MatchFixedString):
            if isinstance(item, str):
                w_item = QListWidgetItem(item)
                w_item.setFlags(flags or Qt.ItemIsSelectable | Qt.ItemIsEditable | Qt.ItemIsEnabled)
                self.addItem(w_item)
            else:
                item.setFlags(flags or Qt.ItemIsSelectable | Qt.ItemIsEditable | Qt.ItemIsEnabled)
                self.addItem(item)

    def remove_item_at_index(self, row: int) -> Optional[QListWidgetItem]:
        """Removes the element at the specified row from this list if it is present. """
        item = self.takeItem(row)
        if item:
            self._items.remove(self._items[row])
        return item

    def del_item(self, item_or_index: Union[int, QModelIndex, QListWidgetItem]) -> Optional[QListWidgetItem]:
        """Removes the specified element from this list if it is present."""
        item = None
        if isinstance(item_or_index, int):
            if 0 <= item_or_index < len(self._items):
                item = self.remove_item_at_index(item_or_index)
        elif isinstance(item_or_index, QListWidgetItem):
            index = self.indexFromItem(item_or_index)
            if index:
                item = self.remove_item_at_index(index.row())
        elif isinstance(item_or_index, QModelIndex):
            item = self.remove_item_at_index(item_or_index.row())

        return item

    def clear(self) -> None:
        """Removes all of the elements from this list."""
        super().clear()
        del self._items[:]

    def size(self) -> int:
        """Returns the number of elements in this list. """
        return self.count()

    def index_of(self, item: str) -> int:
        """Returns the index of the first occurrence of the specified element in this list,
        or -1 if this list does not contain the element."""
        return next((obj for obj in self._items if obj == item), -1)

    def as_list(self) -> List[str]:
        """Returns a list containing all items in this list as a string."""
        return list(map(QListWidgetItem.text, self._items))

    def current_text(self) -> str:
        """Returns the current item's text."""
        return self.currentItem().text()

    def set_context_menu_enable(self, enabled: bool = True) -> None:
        """Whether context menu is enabled or not"""
        self._context_menu_enable = enabled

    def set_editable(self, editable: bool = True) -> None:
        """Set editable property."""
        for item in self._items:
            flags = int(item.flags())
            if editable:
                item.setFlags(flags | Qt.ItemIsEditable)
            else:
                item.setFlags(self.unset_flag(flags, int(Qt.ItemIsEditable)))
        self._editable = editable

    def set_selectable(self, selectable: bool = True) -> None:
        """Set selectable property."""
        for item in self._items:
            flags = int(item.flags())
            if selectable:
                item.setFlags(flags | Qt.ItemIsSelectable)
            else:
                item.setFlags(self.unset_flag(flags, int(Qt.ItemIsSelectable)))
        self._selectable = selectable

    def _context_menu(self) -> None:
        """Display the custom context menu."""
        if self._context_menu_enable:
            ctx_menu = QMenu(self)
            if self._editable:
                ctx_menu.addAction('Add Item', lambda: self.set_item("<new_item>"))
                if not self.is_empty():
                    ctx_menu.addAction('Delete Item', lambda: self.del_item(self.currentIndex().row()))
                    ctx_menu.addSeparator()
                    ctx_menu.addAction('Clear list', self.clear)

            for act in self._custom_menu_actions:
                check_not_none(act)
                check_state(len(act) == 3, f'Invalid custom menu action: {act}')
                check_argument(callable(act[1]), 'The action must be callable')
                if act[2]:
                    ctx_menu.addSeparator()
                ctx_menu.addAction(act[0], act[1])

            ctx_menu.exec_(QCursor.pos())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: hspylib
   @package: hspylib.main.hspylib
      @file: __classpath__.py
   @created: Wed, 8 Jun 2022
    @author: "<B>H</B>ugo <B>S</B>aporetti <B>J</B>unior")"
      @site: "https://github.com/yorevs/hspylib")
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import logging as log
import os
from pathlib import Path
from textwrap import dedent
from typing import Optional

from hspylib.core.exception.exceptions import ResourceNotFoundError, SourceNotFoundError
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.preconditions import check_state


class Classpath(metaclass=Singleton):
    """The classpath tells Python applications where to look in the filesystem for source and resource files."""

    INSTANCE = None

    def __init__(
        self,
        source_root: Optional[Path] = None,
        run_dir: Optional[Path] = None,
        resource_dir: Optional[Path] = None):

        if source_root:
            check_state(source_root.exists(), "source_root must exist")
        if run_dir:
            check_state(run_dir.exists(), "run_dir must exist")

        self._source_root = Path(os.getenv('SOURCE_ROOT', source_root)) or Path(os.curdir)
        self._run_dir = run_dir or Path(os.curdir)
        self._resource_dir = resource_dir or Path(f"{self._source_root}/resources")
        self._log_dir = Path(os.getenv('LOG_DIR', f"{self._run_dir}/log"))

    @classmethod
    def source_root(cls) -> Path:
        """TODO"""
        return cls.INSTANCE._source_root

    @classmethod
    def run_dir(cls) -> Path:
        """TODO"""
        return cls.INSTANCE._run_dir

    @classmethod
    def resource_dir(cls) -> Path:
        """TODO"""
        return cls.INSTANCE._resource_dir

    @classmethod
    def log_dir(cls) -> Path:
        """TODO"""
        return cls.INSTANCE._log_dir

    @classmethod
    def list_resources(cls, directory: str | Path) -> str:
        """TODO"""
        res_str = Classpath.list_files(directory)
        for root, dirs, _ in os.walk(directory):
            for dirname in dirs:
                res_str += '  |-' + dirname + os.linesep
                res_str += cls.list_resources(os.path.join(root, dirname))
        return res_str

    @staticmethod
    def list_files(directory: str | Path) -> str:
        """TODO"""
        res_str = ''
        if (isinstance(directory, str) and os.path.exists(directory)) \
            or (isinstance(directory, Path) and directory.exists()):
            for filename in os.listdir(directory):
                f = os.path.join(directory, filename)
                if os.path.isfile(f):
                    res_str += '    |-' + str(filename) + os.linesep
        return res_str

    @classmethod
    def get_resource_path(cls, resource: str | Path) -> Path:
        """TODO"""
        resource = Path(f'{cls.INSTANCE.resource_dir()}/{str(resource)}')
        if not resource.exists():
            raise ResourceNotFoundError(f'Resource {str(resource)} was not found!')
        return resource

    @classmethod
    def get_source_path(cls, source: str | Path) -> Path:
        """TODO"""
        filepath = Path(f"{cls.INSTANCE.source_root()}/{str(source)}")
        if not filepath.exists():
            raise SourceNotFoundError(f"Source {str(source)} was not found!")
        return filepath

    @classmethod
    def load_envs(cls, prefix: str = 'env', suffix: str = None, load_dir: str = None) -> None:
        """TODO"""
        env_file = f"{load_dir or f'{cls.INSTANCE.source_root()}/env'}/{prefix}{f'-{suffix}' if suffix else ''}.env"
        if os.path.exists(env_file):
            log.debug(f"ENVIRON::Loading environment file '{env_file}'")
            with open(env_file, 'r') as f_env:
                lines = f_env.readlines()
                lines = list(filter(lambda l: l.startswith('export '), filter(None, lines)))
                variables = list(map(lambda x: x.split('=', 1), map(lambda l: l[7:].strip(), lines)))
                for v in variables:
                    log.debug(f"ENVIRON::With environment variable\t'{v[0]}'")
                    os.environ[v[0]] = v[1]
        else:
            log.warning(f"ENVIRON::Environment file '{env_file}' was not found!")

    def __str__(self):
        return dedent(f"""
        |-source-root: {self._source_root}
        |-run-dir: {self.run_dir}
        |-resource-dir: {self._resource_dir}
        |-log-dir: {self._log_dir}
        """) + self.list_resources(self._resource_dir)

    def __repr__(self):
        return str(self)


if __name__ == '__main__':
    Classpath.load_envs(
        suffix='stage',
        load_dir='/Users/hjunior/GIT-Repository/GAP/SBS/LIProductionalisationScripts/astra-db-atp-hydration/env')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.tui.factory
      @file: menu_factory.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from typing import Optional

from hspylib.core.metaclass.singleton import Singleton
from hspylib.modules.cli.tui.menu.menu import Menu
from hspylib.modules.cli.tui.menu.menu_entry import MenuEntry
from hspylib.modules.cli.tui.menu.menu_option import MenuOption


class MenuFactory(metaclass=Singleton):
    """TODO"""

    class MenuBuilder:
        """TODO"""

        def __init__(
            self,
            parent_menu: Optional[Menu] = None,
            menu_title: str = ''):

            self.parent_menu = parent_menu
            self.menu_title = menu_title
            self.menu_options = {}

        def with_option(self, text: str) -> MenuOption:
            """TODO"""
            index = len(self.menu_options)
            option = MenuOption(self, index, text)
            self.menu_options[str(index)] = option
            return option

        def build(self) -> Menu:
            """TODO"""
            return MenuEntry(self.parent_menu, self.menu_options, self.menu_title)

    @staticmethod
    def create(
        parent_menu: Optional[Menu] = None,
        menu_title: str = '') -> MenuBuilder:

        return MenuFactory.MenuBuilder(parent_menu, menu_title)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.icons.emojis.faces
      @file: face_smiling.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.modules.cli.icons.emojis.emojis import Emoji


class FaceSmiling(Emoji):
    """
        Face smiling emojis.
        Codes can be found here:
        - https://unicode.org/emoji/charts/emoji-list.html#face-smiling
    """

    # @formatter:off

    DEFAULT         = '\U0001F600'      # 😀
    BEAMING         = '\U0001F601'      # 😁
    TEARS_OF_JOY    = '\U0001F602'      # 😂
    BIG_EYES        = '\U0001F603'      # 😃
    SMILING_EYES    = '\U0001F604'      # 😄
    SWEAT           = '\U0001F605'      # 😅
    SQUINTING       = '\U0001F606'      # 😆
    HALO            = '\U0001F607'      # 😇
    WINKING         = '\U0001F609'      # 😉
    BLUSHING        = '\U0001F60A'      # 😊
    SLIGHTLY        = '\U0001F642'      # 🙂
    UPSIDE_DOWN     = '\U0001F643'      # 🙃
    ROFL            = '\U0001F923'      # 🤣

    # @formatter:on

    @classmethod
    def demo_emojis(cls) -> None:
        list(map(Emoji.emj_print, cls.values()))


if __name__ == '__main__':
    FaceSmiling.demo_emojis()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.main.hspylib.modules.qt.promotions
      @file: htablemodel.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import collections
from typing import List, Tuple, Type, TypeVar, Union

from PyQt5.QtCore import QAbstractTableModel, QModelIndex, Qt, QVariant
from PyQt5.QtGui import QPalette
from PyQt5.QtWidgets import QTableView

from hspylib.core.tools.collection_filter import CollectionFilter, FilterCondition
from hspylib.core.tools.commons import class_attribute_names, class_attribute_values

T = TypeVar('T')


class HTableModel(QAbstractTableModel):
    """TODO"""

    def __init__(
        self,
        parent: QTableView,
        clazz: Type,
        headers: List[str] = None,
        table_data: List[T] = None,
        cell_alignments: List[Qt.AlignmentFlag] = None,
        max_rows: int = 500):

        QAbstractTableModel.__init__(self, parent)
        self._parent = parent
        self._clazz = clazz
        self._data = collections.deque(maxlen=max_rows)
        self._filters = CollectionFilter()
        self._headers = headers or self._headers_by_entity()
        self._cell_alignments = cell_alignments
        self._parent.setModel(self)
        self.push_data(table_data or [])

    def removeRow(self, row: int, parent: QModelIndex = ...) -> bool:  # pylint: disable=unused-argument
        """TODO"""
        if 0 <= row < len(self._data):
            self.beginRemoveRows(QModelIndex(), row, row + 1)
            del self._data[row]
            self.layoutChanged.emit()
            self.endRemoveRows()
            return True
        return False

    def data(self, index: QModelIndex, role: int = ...) -> QVariant:
        """TODO"""
        if role == Qt.DisplayRole:
            row_dict = self._data[index.row()].__dict__
            value = class_attribute_values(row_dict)[index.column()]
            ret_val = QVariant(str(value if value is not None else ''))
        elif role == Qt.TextAlignmentRole:
            ret_val = QVariant(self._cell_alignments[index.column()]) \
                if self._cell_alignments else QVariant(Qt.AlignLeft | Qt.AlignVCenter)
        elif role == Qt.BackgroundColorRole:
            light_gray = self._parent.palette().color(QPalette.Window).lighter(100)
            ret_val = QVariant(light_gray if index.row() % 2 != 0 else '')
        else:
            ret_val = None

        return ret_val or QVariant()

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = ...) -> QVariant:
        """TODO"""
        ret_val = None
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            ret_val = QVariant(self._headers[section].upper()) \
                if len(self._headers) >= section else QVariant('-')
        elif orientation == Qt.Vertical and role == Qt.DisplayRole:
            ret_val = QVariant(str(section))

        return ret_val if ret_val else QVariant()

    def rowCount(self, parent: QModelIndex = ...) -> int:  # pylint: disable=unused-argument
        """TODO"""
        return len(self._data) if self._data else 0

    def columnCount(self, parent: QModelIndex = ...) -> int:  # pylint: disable=unused-argument
        """TODO"""
        return len(self._headers) if self._data else 0

    def sort(self, column: int, order: Qt.SortOrder = ...) -> None:
        """TODO"""
        keys = class_attribute_names(self._clazz)
        self._data = sorted(self._data, key=lambda x: getattr(x, keys[column]), reverse=bool(order))
        self.layoutChanged.emit()

    def append(self, data: T):
        """TODO"""
        if data and not self._filters.should_filter(data):
            self.beginInsertRows(QModelIndex(), self.rowCount(), self.rowCount() + 1)
            self._data.append(data)
            self.endInsertRows()
            self.layoutChanged.emit()

    def apply_filter(
        self,
        name: str,
        el_name: str,
        condition: FilterCondition,
        el_value: Union[int, str, bool, float]) -> None:
        """TODO"""
        self._filters.apply_filter(name, el_name, condition, el_value)

    def filter(self) -> None:
        """TODO"""
        self._data = self._filters.filter(list(self._data))

    def row(self, index: QModelIndex) -> T:
        """TODO"""
        return self._data[index.row()]

    def column(self, index: QModelIndex) -> T:
        """TODO"""
        row = self.row(index)
        col_name = str(list(vars(row))[index.column()])
        return getattr(row, col_name)

    def push_data(self, data: Union[List[T], T]) -> None:
        """TODO"""
        if data:
            if isinstance(data, list):
                list(map(self.append, data))
            else:
                self.append(data)
            self.layoutChanged.emit()

    def clear(self):
        """TODO"""
        self._data.clear()
        self.layoutChanged.emit()

    def is_empty(self) -> bool:
        """TODO"""
        return len(self._data) == 0

    def refresh_data(self) -> None:
        """TODO"""
        self._data = self._filters.filter(list(self._data))
        self.layoutChanged.emit()

    def remove_rows(self, rows: List[QModelIndex]):
        """TODO"""
        # Because we are using deque, we need to sort DESC to avoid deleting wrong indexes
        for row in sorted(rows, reverse=True):
            self.removeRow(row.row())

    def _headers_by_entity(self) -> List[str]:
        """TODO"""
        attributes = class_attribute_names(self._clazz)
        return [str(x).capitalize() for x in attributes]

    def selected_data(self) -> Tuple[List[QModelIndex], List[T]]:
        """TODO"""
        sel_model = self._parent.selectionModel()
        if sel_model:
            sel_indexes = sel_model.selectedIndexes()
            sel_rows = set([idx.row() for idx in sel_indexes])
            return sel_indexes, [self._data[row] for row in sel_rows]

    def filters(self) -> CollectionFilter:
        """TODO"""
        return self._filters



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.config
      @file: properties.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import logging as log
import os
import re
from collections import defaultdict
from configparser import ConfigParser
from typing import Any, Callable, Iterator, List, Optional, Type

import yaml

from hspylib.core.enums.charset import Charset
from hspylib.core.tools.commons import run_dir, touch_file
from hspylib.core.tools.dict_tools import flatten_dict


class Properties:
    """The Properties class represents a persistent set of properties. Each key and its corresponding value in the
    property list is a string."""

    _default_name = 'application'
    _default_ext = '.properties'

    @staticmethod
    def environ_name(property_name: str) -> str:
        """Retrieve the environment name of the specified property name
        :param property_name: the name of the property using space, dot or dash notations
        """
        return re.sub('[ -.]', '_', property_name).upper()

    @staticmethod
    def _read_properties(all_lines: List[str]) -> dict:
        """Reads a property list (key and element pairs) from the input list."""
        return {
            p[0].strip(): p[1].strip() for p in [
                p.split('=', 1) for p in list(
                    filter(lambda l: re.match(r'[a-zA-Z]([.\\-]|\w)* *= *.+', l), all_lines)
                )
            ]
        }

    @staticmethod
    def _read_cfg_or_ini(all_lines: List[str]):
        """Reads a cfg or ini list (key and element pairs) from the input list."""
        string = os.linesep.join(all_lines)
        all_cfgs = {}
        cfg = ConfigParser()
        cfg.read_string(string)
        for section in cfg.sections():
            all_cfgs.update(dict(cfg.items(section)))
        return all_cfgs

    def __init__(self, filename: Optional[str] = None, profile: Optional[str] = None, load_dir: Optional[str] = None):
        self._filename, self._extension = os.path.splitext(
            filename if filename else f'{self._default_name}{self._default_ext}')
        self._profile = profile if profile else os.environ.get('ACTIVE_PROFILE', '')
        self._properties = defaultdict()
        self._load(load_dir or f'{run_dir()}/resources')

    def __str__(self) -> str:
        str_val = ''
        for key, value in self._properties.items():
            str_val += '{}{}={}'.format('\n' if str_val else '', key, value)
        return str_val

    def __repr__(self) -> str:
        return str(self)

    def __getitem__(self, item: str) -> Optional[Any]:
        return self.get(item)

    def __iter__(self) -> Iterator:
        return self._properties.__iter__()

    def __len__(self) -> int:
        """Retrieve the amount of properties"""
        return len(self._properties)

    def get(self, prop_name: str, prop_type: Type | Callable = str, default: Optional[str] = None) -> Optional[Any]:
        """Retrieve a property specified by property and cast to the proper type. If the property is not found,
        return the default value."""
        try:
            value = self._get(prop_name)
            return prop_type(value) if value else None
        except TypeError:
            log.warning(f"Unable to cast property '{prop_name}' into '{prop_type}'")
            return default

    @property
    def values(self) -> List[Any]:
        """Retrieve all values for all properties"""
        return list(self._properties.values())

    @property
    def keys(self) -> List[str]:
        """Retrieve all values for all properties"""
        return list(self._properties.keys())

    @property
    def size(self) -> int:
        """Retrieve the amount of properties actually store."""
        return len(self._properties)

    def _get(self, key: str, default_value: Any = None) -> Optional[Any]:
        """Get a property value as string or default_value if the property was not found"""
        value = os.environ.get(self.environ_name(key), None)
        if value:
            return value
        return self._properties[key] if key in self._properties else default_value

    def _load(self, load_dir: str) -> None:
        """Read all properties from the file"""
        filepath = self._build_path(load_dir)
        if os.path.exists(filepath):
            return self._parse(filepath)

        raise FileNotFoundError(
            f'File "{filepath}" does not exist')

    def _build_path(self, load_dir: str) -> str:
        """Find the proper path for the properties file"""
        if self._profile:
            filepath = f'{load_dir}/{self._filename}-{self._profile}{self._extension}'
        else:
            filepath = f'{load_dir}/{self._filename}{self._extension}'

        return filepath

    def _parse(self, filepath: str) -> None:
        """Parse the properties file according to it's extension"""
        if not os.path.isfile(filepath):
            touch_file(filepath)
        ext = self._extension.lower()
        with open(filepath, encoding=str(Charset.UTF_8)) as fh_props:
            if ext in ['.ini', '.cfg']:
                all_lines = list(map(str.strip, filter(None, fh_props.readlines())))
                self._properties.update(self._read_cfg_or_ini(all_lines))
            elif ext == '.properties':
                all_lines = list(map(str.strip, filter(None, fh_props.readlines())))
                self._properties.update(self._read_properties(all_lines))
            elif ext in ['.yml', '.yaml']:
                self._properties.update(flatten_dict(yaml.safe_load(fh_props)))
            else:
                raise NotImplementedError(f"Extension {ext} is not supported")
        log.debug('Successfully loaded %d properties from: %s', len(self._properties), filepath)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.hspylib.core.datasource
      @file: crud_service.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from typing import Generic, List, Optional, TypeVar

from hspylib.core.datasource.crud_entity import CrudEntity
from hspylib.core.datasource.crud_repository import CrudRepository
from hspylib.core.datasource.identity import Identity
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.tools.namespace import Namespace

E = TypeVar('E', bound=CrudEntity)
R = TypeVar('R', bound=CrudRepository)


class CrudService(Generic[R, E], metaclass=Singleton):

    def __init__(self, repository: R):
        self._repository = repository

    @property
    def repository(self) -> R:
        return self._repository

    def list(self, filters: Optional[Namespace] = None, order_bys: Optional[List[str]] = None) -> List[E]:
        return self.repository.find_all(filters=filters, order_bys=order_bys)

    def remove(self, entity: E) -> None:
        self.repository.delete(entity)

    def get(self, identity: Identity) -> Optional[E]:
        return self.repository.find_by_id(identity)

    def save(self, entity: E) -> None:
        self.repository.save(entity)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.icons.font_awesome
      @file: awesome.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import re
import struct
from typing import Union

from hspylib.core.enums.enumeration import Enumeration
from hspylib.core.tools.commons import sysout
from hspylib.core.tools.text_tools import eol


class Awesome(Enumeration):
    """
        Font awesome codes
        Full list of font awesome icons can be found here:
          - https://fontawesome.com/cheatsheet?from=io
    """

    @staticmethod
    def compose(*values: 'Awesome') -> str:
        result = ''
        for v in values:
            result += str(v.value)
        return result

    @staticmethod
    def print_unicode(uni_code: Union[str, int], end: str = '') -> None:
        """TODO"""
        if isinstance(uni_code, str) and re.match(r"^[a-fA-F0-9]{1,4}$", uni_code):
            hex_val = bytes.decode(struct.pack("!I", int(uni_code.zfill(4), 16)), 'utf_32_be')
            sysout(f"{hex_val:2s}", end=end)
        elif isinstance(uni_code, int):
            hex_val = bytes.decode(struct.pack("!I", uni_code), 'utf_32_be')
            sysout(f"{hex_val:2s}", end=end)
        else:
            raise TypeError(f'Invalid unicode value: {uni_code}')

    @staticmethod
    def demo_unicodes(split_columns: int = 16, fa_start: int = 0xF000, fa_end: int = 0xFD50) -> None:
        """TODO"""
        fa_range = fa_start, fa_end  # Font awesome range unicodes
        st_base = [f'{hex(x)[2:]}' for x in range(*fa_range)]
        for n, h in enumerate(st_base):
            Awesome.print_unicode(h)
            sysout(f'{h:04}', end=eol(n, split_columns))

    @classmethod
    def demo_icons(cls, split_columns: int = 16) -> None:
        """TODO"""
        for i, n in enumerate(cls.values()):
            sysout(f'{n:2}', end=eol(i, split_columns))

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self):
        return str(self)

    def placeholder(self) -> str:
        """TODO"""
        return f":{self.name}:"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.enum
      @file: http_method.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.core.enums.enumeration import Enumeration


class HttpMethod(Enumeration):
    """Associates the name of a HTTP method with an enumeration."""

    # @formatter:off

    # The GET method requests a representation of the specified resource. Requests using GET should only retrieve data.
    GET         = 'get'

    # The HEAD method asks for a response identical to a GET request, but without the response body.
    HEAD        = 'head'

    # The POST method submits an entity to the specified resource, often causing a change in state or side effects
    # on the server.
    POST        = 'post'

    # The PUT method replaces all current representations of the target resource with the request payload.
    PUT         = 'put'

    # The DELETE method deletes the specified resource.
    DELETE      = 'delete'

    # The CONNECT method establishes a tunnel to the server identified by the target resource.
    CONNECT     = 'connect'

    # The OPTIONS method describes the communication options for the target resource.
    OPTIONS     = 'options'

    # The TRACE method performs a message loop-back test along the path to the target resource.
    TRACE       = 'trace'

    # The PATCH method applies partial modifications to a resource.
    PATCH       = 'patch'

    # @formatter:on
